//  t4.1 PIN input  and image packet test
//  New CameraBB with new pinouts for wired breadboard
//  MJB  10/27/2020

#include <stdint.h>
#include <Wire.h>
#include <SD.h>

#define FRAME_WIDTH 640l
#define FRAME_HEIGHT 480l

#define FB_WIDTH 640l
#define FB_HEIGHT 480l
#define FB_COUNT 2  // two frames in EXTMEM

#define FRAMEBYTES (FRAME_WIDTH * FRAME_HEIGHT *2)
#define FRAMEBUFFBYTES (FB_WIDTH *FB_HEIGHT *2)

uint8_t fb1[FRAMEBUFFBYTES] EXTMEM;
uint8_t fb2[FRAMEBUFFBYTES] EXTMEM;

uint32_t masterfreq = 24;
volatile uint32_t last_csistat, csiIrqCount;
volatile uint32_t fb1count, fb2count;
volatile uint32_t bswitcherrors;
uint32_t framecount;


const char compileTime [] = " Compiled on " __DATE__ " " __TIME__;

const int ledpin    = 13;
const int imarkpin = 32;
const int pinCamReset = 14;


#define  LEDON digitalWriteFast(ledpin, HIGH); // Also marks IRQ handler timing
#define  LEDOFF digitalWriteFast(ledpin, LOW);
#define  LEDTOGGLE digitalToggleFast(ledpin);

#define  IMARKHI digitalWriteFast(imarkpin, HIGH); // Also marks IRQ handler timing
#define  IMARKLO digitalWriteFast(imarkpin, LOW);
#define  IMARKTOGGLE digitalToggleFast(imarkpin);

#define BMPIMAGEOFFSET 66
const char bmp_header[BMPIMAGEOFFSET] PROGMEM =
{
  0x42, 0x4D, 0x36, 0x58, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x42, 0x00, 0x00, 0x00, 0x28, 0x00,
  0x00, 0x00, 0x40, 0x01, 0x00, 0x00, 0xF0, 0x00, 0x00, 0x00, 0x01, 0x00, 0x10, 0x00, 0x03, 0x00,
  0x00, 0x00, 0x00, 0x58, 0x02, 0x00, 0xC4, 0x0E, 0x00, 0x00, 0xC4, 0x0E, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, 0x00, 0x00, 0xE0, 0x07, 0x00, 0x00, 0x1F, 0x00,
  0x00, 0x00
};

# pragma pack (2)
typedef struct  tBMPHDR565{
  uint16_t  bfType = 0x4d42;   //'bm';
  uint32_t  bfSize = 614466;// 614400 pixels + 66 header 
  uint16_t  bfReserved1 = 0;
  uint16_t  bfReserved2 = 0;
  uint32_t  bfOffBits =  66; // 14 bytes to here
  uint32_t  biSize = 40;
  int32_t   biWidth = 640;
  int32_t   biHeight = -480;  // windows wants negative for top-down image
  int16_t   biPlanes = 1;
  uint16_t  biBitCount = 16 ;
  uint32_t  biCompression = 3;  // bitfields used
  uint32_t  biSizeImage = 614400;  // 640 * 480 * 2
  int32_t   biXPelsPerMeter = 0;
  int32_t   biYPelsPerMeter = 0;
  uint32_t  biClrUsed  = 0;
  uint32_t  biClrImportant = 0;// 54 bytes
  uint32_t  rmask = 0x0000F800;
  uint32_t  gmask = 0x000007E0; 
  uint32_t  bmask = 0x0000001F;  //66 bytes
} BMPHDR565;

tBMPHDR565 fileheader;
File bmpfile;

void setup() {
  Serial.begin(9600);
  delay(200);
  Wire.begin();

  delay(100);
  pinMode(ledpin, OUTPUT);
  pinMode(imarkpin, OUTPUT);
  pinMode(pinCamReset, OUTPUT);

  delay(10);
  memset(&fb1, 0, FRAMEBUFFBYTES * 2);

  digitalWriteFast(pinCamReset, LOW);
  delay(10);
  digitalWriteFast(pinCamReset, HIGH);  // subsequent resets via SCB


  Serial.printf("\n\nOV7670 Camera Test 3 %s\n", compileTime);
  cameraBegin(640, 480, fb1, fb2);

}

void loop() {
  static uint32_t lastfc;
  static uint32_t fc = 0;
  char ch;
  if (Serial.available()) {
    ch = Serial.read();
    if (ch == 's') CMSI();
    if (ch == 'c') CMCS();
    if (ch == 'r') CMCR();
    if (ch == 'f') CMGF();
    if (ch == 'x') send_raw();
    if (ch == 'd') save_image_SD();
  }

  //  This extends frame count from 16 to 32 bits
  fc  = CSI_CSICR3 >> 16; // get the 16-bit frame count
 
  framecount += (fc - lastfc);
  lastfc = fc;

  switch(ch) {
      case 0x30:
        Serial.println(F("ACK CMD CAM start single shoot. END"));
        send_image();
          Serial.println(F("READY. END"));
        break;
      default:
        break;
  }

}

// Show information
void CMSI(void) {

  Serial.printf("\n\nOV7670 Camera Test 3 %s\n", compileTime);
  Serial.printf("buff1 at %p   buff2 at %p\n", &fb1, &fb2);
  Serial.printf("Camera frame count = %lu \n", framecount);
  Serial.printf("Base Address Change errors: %lu\n", bswitcherrors);
  Serial.printf("csistat: %08X, csiIrqCount: %lu\n", CSI_CSISR, csiIrqCount);
  Serial.printf("fb1 count: %6lu   fb2 count %6lu\n",fb1count,fb2count);
}

void CMCS(void) {  // Show CSI registers
  print_csi_registers();
}



// Show Camera Registers
void CMCR(void) {
  uint8_t regs[200];
  cameraReadAll(regs);
  cameraShowAll(regs);
}


// Get a frame of data
void CMGF(void) {
  uint32_t imagesize, fbc;
  imagesize = (FRAME_WIDTH * FRAME_HEIGHT * 2);
  fbc = fb2count;
  while (fbc == fb2count); // wait until a new frame in fb2 is finished
  // Don't try sending unless your program can handle 614400 bytes in a hurry!
  //Serial.write(fb2,magesize);
  Serial.printf("fb2 ready for processing %lu nytes.", imagesize);
}

void send_image() {
  uint32_t imagesize, fbc;
  imagesize = (FRAME_WIDTH * FRAME_HEIGHT * 2);
  fbc = fb2count;
  while (fbc == fb2count); // wait until a new frame in fb2 is finished

  
    Serial.write(0xFF);
    Serial.write(0xAA);
  
    for (int temp = 0; temp < BMPIMAGEOFFSET; temp++)
    {
      Serial.write(pgm_read_byte(&bmp_header[temp]));
    }
    //Serial.write((const uint8_t *)&fileheader, sizeof(fileheader));

    Serial.write(fb2, imagesize);

    Serial.write(0xBB);
    Serial.write(0xCC);
}

void send_raw() {
    uint32_t imagesize, fbc;
  imagesize = (FRAME_WIDTH * FRAME_HEIGHT * 2);
  fbc = fb2count;
  while (fbc == fb2count); // wait until a new frame in fb2 is finished
    Serial.write(fb2, imagesize);

}

void save_image_SD() {
  if (!SD.begin(BUILTIN_SDCARD)) {
    Serial.println("initialization failed!");
    while (1){
        LEDON; delay(100);
        LEDOFF; delay(100);
      }
  }
  Serial.println("initialization done.");

  delay(100);

  uint32_t imagesize, fbc;
  imagesize = (FRAME_WIDTH * FRAME_HEIGHT * 2);
  fbc = fb2count;
  while (fbc == fb2count); // wait until a new frame in fb2 is finished
  // Don't try sending unless your program can handle 614400 bytes in a hurry!
  //Serial.write(fb2,magesize);
  Serial.printf("fb2 ready to save %lu bytes.", imagesize);
  Serial.println("Saving OV7670.bmp ");
  if (SD.exists("OV7670.bmp")) {
    // delete the file:
    Serial.println("Removing OV7670.bmp...");
    SD.remove("OV7670.bmp");
  }
  bmpfile = SD.open("OV7670.bmp", FILE_WRITE);
  bmpfile.write((const uint8_t *)&fileheader, sizeof(fileheader));
  bmpfile.write(fb2, imagesize);
  bmpfile.close();
  Serial.println("File saved to SD card.");
}
